@@ -3,9 +3,10 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  2002-2022, University of Amsterdam
+    Copyright (c)  2002-2024, University of Amsterdam
                               VU University Amsterdam
                               CWI, Amsterdam
+                              SWI-Prolog Solutions b.v.
     All rights reserved.
 
     Redistribution and use in source and binary forms, with or without
@@ -58,6 +59,7 @@
 :- use_module(http_path).
 
 :- autoload(library(uri), [uri_resolve/3]).
+:- autoload(library(aggregate), [aggregate_all/3]).
 
 :- predicate_options(http_server/2, 2,
                      [ port(any),
@@ -152,7 +154,8 @@ self-signed SSL certificate.
 %     * tcp_socket(+Socket)
 %     If provided, use this socket instead of the creating one and
 %     binding it to an address.  The socket must be bound to an
-%     address.
+%     address.  Note that this also allows binding an HTTP server to
+%     a Unix domain socket (``AF_UNIX``).  See socket_create/2.
 %
 %     * workers(+Count)
 %     Determine the number of worker threads.  Default is 5.  This
@@ -242,7 +245,7 @@ address_domain(_, Domain) =>
 %!  make_socket(+Address, :OptionsIn, -OptionsOut) is det.
 %
 %   Create the HTTP server socket and  worker pool queue. OptionsOut
-%   is quaranteed to hold the option queue(QueueId).
+%   is guaranteed to hold the option queue(QueueId).
 %
 %   @arg   OptionsIn   is   qualified   to     allow   passing   the
 %   module-sensitive ssl option argument.
@@ -390,24 +393,27 @@ server_property(start_time(Time), Port) :-
     current_server(Port, _, _, _, _, Time).
 
 
-%!  http_workers(+Port, -Workers) is det.
+%!  http_workers(?Port, -Workers) is nondet.
 %!  http_workers(+Port, +Workers:int) is det.
 %
-%   Query or set the number of workers  for the server at this port.
-%   The number of workers is dynamically   modified. Setting it to 1
-%   (one) can be used to profile the worker using tprofile/1.
+%   Query or set the number of workers for  the server at this port. The
+%   number of workers is dynamically modified. Setting it to 1 (one) can
+%   be used to profile the worker using tprofile/1.
+%
+%   @see library(http/http_dyn_workers) implements dynamic management of
+%   the worker pool depending on usage.
 
 http_workers(Port, Workers) :-
-    must_be(ground, Port),
-    current_server(Port, _, _, Queue, _, _),
+    integer(Workers),
     !,
-    (   integer(Workers)
+    must_be(ground, Port),
+    (   current_server(Port, _, _, Queue, _, _)
     ->  resize_pool(Queue, Workers)
-    ;   findall(W, queue_worker(Queue, W), WorkerIDs),
-        length(WorkerIDs, Workers)
+    ;   existence_error(http_server, Port)
     ).
-http_workers(Port, _) :-
-    existence_error(http_server, Port).
+http_workers(Port, Workers) :-
+    current_server(Port, _, _, Queue, _, _),
+    aggregate_all(count, queue_worker(Queue, _Worker), Workers).
 
 
 %!  http_add_worker(+Port, +Options) is det.
@@ -491,8 +497,6 @@ send_to_worker(Queue, Client, Goal, Peer) :-
     thread_send_message(Queue, tcp_client(Client, Goal, Peer)).
 
 accept_rethrow_error(http_stop(_)).
-accept_rethrow_error('$aborted').
-
 
 %!  close_server_socket(+Options)
 %
@@ -819,6 +823,8 @@ done_worker :-                                  % received quit(Sender)
 
 done_status_message_level(true, silent) :- !.
 done_status_message_level(exception('$aborted'), silent) :- !.
+done_status_message_level(exception(unwind(abort)), silent) :- !.
+done_status_message_level(exception(unwind(halt(_))), silent) :- !.
 done_status_message_level(_, informational).
 
 
@@ -843,6 +849,7 @@ recreate_worker(exception(Error), Queue) :-
     create_workers(1, 1, Queue, AliasBase, Options).
 
 recreate_on_error('$aborted').
+recreate_on_error(unwind(abort)).
 recreate_on_error(time_limit_exceeded).
 
 %!  thread_httpd:message_level(+Exception, -Level)
@@ -1056,7 +1063,7 @@ long(_, long).
 %!  update_rate_mma(!State, -MMA) is det.
 %
 %   Implement _Modified Moving  Average_  computing   the  average  time
-%   between requests as an exponential moving averate with alpha=1/N.
+%   between requests as an exponential moving average with alpha=1/N.
 %
 %   @arg Resolution is the time resolution  in 1/Resolution seconds. All
 %   storage is done in integers to avoid  the need for stack freezing in
